from typing import Any, Iterable

import sqlalchemy


def getattr_fallback(
    object: Any,
    name: str,
    *fallbacks: Iterable[str],
    default: Any = None,
) -> Any:
    attr = getattr(object, name, None)
    if attr is not None:
        return attr

    for fallback in fallbacks:
        attr = getattr(object, fallback, None)
        if attr is not None:
            return attr

    return default


def is_sa_mapped(obj):
    try:
        sqlalchemy.orm.object_mapper(obj)
    except sqlalchemy.orm.exc.UnmappedInstanceError:
        return False
    return True



import pytest


@pytest.mark.parametrize(
    "examples_client", [["complete_db", "create_app"]], indirect=True
)
def test_complete(examples_client):
    r = examples_client.post(
        "/v1/authors",
        json=dict(
            name="asdf",
        ),
    )
    assert r.status_code == 201
    assert r.json["name"] == "asdf"

    r = examples_client.post(
        "/v1/authors",
        json=dict(
            name="qwer",
        ),
    )
    assert r.status_code == 201
    assert r.json["name"] == "qwer"

    r = examples_client.get("/v1/authors")
    assert r.status_code == 200
    assert len(r.json["items"]) == 2

    r = examples_client.post(
        "/v1/books",
        json=dict(
            author_id=1,
            title="asdf 1",
            year=1982,
            description="asdf1 asdf1 asdf1",
        ),
    )
    assert r.status_code == 201

    r = examples_client.post(
        "/v1/books",
        json=dict(
            author_id=1,
            title="asdf 2",
            year=1982,
            description="asdf2 asdf2 asdf2",
        ),
    )
    assert r.status_code == 201

    r = examples_client.post(
        "/v1/books",
        json=dict(
            author_id=2,
            title="asdf 3",
            year=1982,
            description="asdf3 asdf3 asdf3",
        ),
    )
    assert r.status_code == 201

    r = examples_client.post(
        "/v1/books",
        json=dict(
            year=1982,
        ),
    )
    assert r.status_code == 400
    assert r.json["source"] == "body"

    r = examples_client.get("/v1/books")
    assert r.status_code == 200
    assert len(r.json["items"]) == 3

    r = examples_client.get("/v1/authors/1/books")
    assert r.status_code == 200
    assert len(r.json["items"]) == 2

    r = examples_client.get("/v1/books/1")
    assert r.status_code == 200
    assert r.json["id"] == 1
    assert r.json["title"] == "asdf 1"

    r = examples_client.get("/v1/books/asdf")
    assert r.status_code == 400
    assert r.json["source"] == "kwargs"

    r = examples_client.patch(
        "/v1/books/1",
        json=dict(
            title="something else",
        ),
    )
    assert r.status_code == 200
    assert r.json["title"] == "something else"

    r = examples_client.get("/v1/books/1")
    assert r.status_code == 200
    assert r.json["id"] == 1
    assert r.json["title"] == "something else"
    assert r.json["description"] == "asdf1 asdf1 asdf1"

    r = examples_client.put(
        "/v1/books/1",
        json=dict(
            title="zxcv",
        ),
    )
    assert r.status_code == 400

    r = examples_client.put(
        "/v1/books/1",
        json=dict(
            title="zxcv",
            year=1983,
            description="zxcv zxcv zxcv",
        ),
    )
    assert r.status_code == 200

    r = examples_client.get("/v1/books/1")
    assert r.status_code == 200
    assert r.json["id"] == 1
    assert r.json["title"] == "zxcv"
    assert r.json["year"] == 1983
    assert r.json["description"] == "zxcv zxcv zxcv"

    r = examples_client.delete("/v1/books/1")
    assert r.status_code == 200

    r = examples_client.get("/v1/books/1")
    assert r.status_code == 404



import importlib
import sys

import pytest


@pytest.fixture(scope="function")
def examples_client(request):
    name = request.param[0]
    app_or_create_app = request.param[1]

    sys.path.append("examples")
    module = importlib.import_module(name)

    if app_or_create_app == "app":
        app = module.app
    elif app_or_create_app == "create_app":
        app = module.create_app(conf="test")

    with app.app_context():
        with app.test_client() as client:
            yield client

    del sys.modules[name]
    sys.path.remove("examples")



import pytest


@pytest.mark.parametrize("examples_client", [["minimal_db", "app"]], indirect=True)
def test_index(examples_client):
    r = examples_client.post("/books", json=dict(title="asdf"))
    assert r.json["title"] == "asdf"

    r = examples_client.get("/books")
    assert len(r.json["items"]) == 1
    assert r.json["items"][0]["id"] == 1
    assert r.json["items"][0]["title"] == "asdf"



import http

import flask
import marshmallow


class ApiError(Exception):
    def __init__(self, status_code, **kwargs):
        self.status_code = status_code
        self.kwargs = kwargs

    def get_error_name(self):
        try:
            return http.HTTPStatus(self.status_code).name
        except ValueError:
            return "UNKNOWN"

    def to_dict(self):
        data = dict(
            status=self.status_code,
            name=self.get_error_name(),
            **self.kwargs,
        )
        if isinstance(self.__cause__, marshmallow.ValidationError):
            data["messages"] = self.__cause__.messages
        return data


def api_error_handler(_, e):
    return flask.jsonify(e.to_dict()), e.status_code



from . import errors, views


class ApiFramework:
    List = views.List
    Read = views.Read
    Create = views.Create
    Update = views.Update
    Delete = views.Delete
    ApiError = errors.ApiError
    api_error_handler = errors.api_error_handler

    def __init__(self, app=None, *args, **kwargs):
        if app is not None:
            self.init_app(app, *args, **kwargs)

    def init_app(self, app, db=None):
        self.db = db
        app.extensions["api-framework"] = self



import flask
import flask.views
import marshmallow
import sqlalchemy

from .errors import ApiError
from .helpers import getattr_fallback, is_sa_mapped


class BaseViewType(type):
    def __init__(cls, name, bases, d):
        super().__init__(name, bases, d)

        if "methods" not in d:
            methods = set()

            for base in bases:
                if getattr(base, "methods", None):
                    methods.update(base.methods)

            for key in flask.views.http_method_funcs:
                if hasattr(cls, f"handle_{key}"):
                    methods.add(key.upper())

            cls.methods = methods

        actions = ["list", "read", "create", "update", "delete"]
        schema_names = {
            "{action}_kwargs": ["kwargs"],
            "{action}_args": ["args"],
            "{action}_request_body": ["request_body", "{action}_body", "body"],
            "{action}_response_body": ["response_body", "{action}_body", "body"],
        }
        for _ in bases:
            for action in actions:
                for name, fallbacks in schema_names.items():
                    name = f"{name}_schema".format(action=action)
                    fallbacks = [
                        f"{fallback}_schema".format(action=action)
                        for fallback in fallbacks
                    ]
                    attr = getattr_fallback(cls, name, *fallbacks)
                    if attr is not None:
                        setattr(cls, name, attr)


class BaseView(flask.views.View, metaclass=BaseViewType):
    kwargs_schema = None
    args_schema = None
    body_schema = None

    loaded_kwargs = None
    loaded_args = None

    @property
    def db(self):
        return flask.current_app.extensions["api-framework"].db

    def dispatch_request(self, *args, **kwargs):
        try:
            method = flask.request.method.lower()
            handler = getattr(self, f"handle_{method}", None)
            return flask.current_app.ensure_sync(handler)(*args, **kwargs)
        except marshmallow.exceptions.ValidationError as e:
            raise ApiError(status_code=400, source="body") from e

    def load_kwargs_schema(self):
        self.loaded_kwargs = self.kwargs_schema.load(data=flask.request.view_args)

    def load_args_schema(self):
        self.loaded_args = self.args_schema.load(data=flask.request.args)

    def handle_kwargs(self, schema):
        try:
            if schema is not None:
                self.kwargs_schema = schema
                self.load_kwargs_schema()
        except marshmallow.exceptions.ValidationError as e:
            raise ApiError(status_code=400, source="kwargs") from e
        if (
            is_sa_mapped(self.loaded_kwargs)
            and sqlalchemy.inspect(self.loaded_kwargs).transient
        ):
            raise ApiError(status_code=404)

    def handle_args(self, schema):
        try:
            if schema is not None:
                self.args_schema = schema
                self.load_args_schema()
        except marshmallow.exceptions.ValidationError as e:
            raise ApiError(status_code=400, source="args") from e
        if (
            is_sa_mapped(self.loaded_args)
            and sqlalchemy.inspect(self.loaded_args).transient
        ):
            raise ApiError(status_code=404)

    def get_schema_success_data(
        self,
        schema,
        data,
        status_code=200,
        data_wrapper=None,
        many=False,
    ):
        def noop_data_wrapper(x):
            return x

        if data_wrapper is None:
            data_wrapper = noop_data_wrapper
        if schema is not None:
            return (
                flask.jsonify(data_wrapper(schema.dump(data, many=many))),
                status_code,
            )
        elif schema is None and data is not None:
            return flask.jsonify(data_wrapper(data)), status_code
        else:
            return "", 204


class List(BaseView):
    list_kwargs_schema = None
    list_args_schema = None
    list_body_schema = None
    list_response_body_schema = None
    response_body_schema = None

    instances = None

    def handle_get(self, *args, **kwargs):
        self.handle_kwargs(self.list_kwargs_schema)
        self.handle_args(self.list_args_schema)
        self.instances = self.get_instances()
        return self.get_list_response()

    def get_instances(self):
        return []

    def get_list_response(self):
        return self.get_schema_success_data(
            schema=self.list_response_body_schema,
            data=self.instances,
            data_wrapper=lambda data: dict(items=data),
            many=True,
        )


class Read(BaseView):
    read_kwargs_schema = None
    read_args_schema = None
    read_body_schema = None
    read_response_body_schema = None
    response_body_schema = None

    instance = None

    def handle_get(self, *args, **kwargs):
        self.handle_kwargs(self.read_kwargs_schema)
        self.handle_args(self.read_args_schema)
        self.instance = self.get_instance()
        return self.get_read_response()

    def get_instance(self):
        return self.loaded_kwargs

    def get_read_response(self):
        return self.get_schema_success_data(
            schema=self.read_response_body_schema,
            data=self.instance,
        )


class Create(BaseView):
    create_kwargs_schema = None
    create_args_schema = None
    create_body_schema = None
    create_request_body_schema = None
    create_response_body_schema = None
    request_body_schema = None
    response_body_schema = None

    loaded_body = None
    instance = None

    def handle_post(self, *args, **kwargs):
        self.handle_kwargs(self.create_kwargs_schema)
        self.handle_args(self.create_args_schema)
        if self.create_request_body_schema is not None:
            self.body_schema = self.create_request_body_schema
            self.load_create_request_body_schema()
        self.instance = self.create()
        self.db.session.commit()
        return self.get_create_response()

    def load_create_request_body_schema(self):
        self.loaded_body = self.create_request_body_schema.load(data=flask.request.json)

    def create(self):
        if is_sa_mapped(self.loaded_body):
            self.db.session.add(self.loaded_body)
        return self.loaded_body

    def get_create_response(self):
        return self.get_schema_success_data(
            schema=self.create_response_body_schema,
            data=self.instance,
            status_code=201,
        )


class Update(BaseView):
    update_kwargs_schema = None
    update_args_schema = None
    update_body_schema = None
    update_request_body_schema = None
    update_response_body_schema = None
    request_body_schema = None
    response_body_schema = None

    loaded_body = None
    instance = None

    def handle_patch(self, *args, **kwargs):
        return self.handle_update(*args, partial=True, **kwargs)

    def handle_put(self, *args, **kwargs):
        return self.handle_update(*args, partial=False, **kwargs)

    def handle_update(self, *args, partial, **kwargs):
        self.handle_kwargs(self.update_kwargs_schema)
        self.handle_args(self.update_args_schema)
        self.instance = self.get_instance()
        if self.update_request_body_schema is not None:
            self.body_schema = self.load_update_request_body_schema
            self.load_update_request_body_schema(partial=partial)
        self.instance = self.update()
        self.db.session.commit()
        return self.get_update_response()

    def get_instance(self):
        return self.loaded_kwargs

    def load_update_request_body_schema(self, partial):
        kwargs = dict(
            data=flask.request.json,
            partial=partial,
        )
        if self.get_schema_load_instance(self.update_request_body_schema):
            kwargs["instance"] = self.instance
        self.loaded_body = self.update_request_body_schema.load(**kwargs)

    def get_schema_load_instance(self, schema):
        try:
            return schema.Meta.load_instance
        except AttributeError:
            return False

    def update(self):
        return self.instance

    def get_update_response(self):
        return self.get_schema_success_data(
            schema=self.update_response_body_schema,
            data=self.instance,
        )


class Delete(BaseView):
    delete_kwargs_schema = None
    delete_args_schema = None
    delete_body_schema = None
    delete_response_body_schema = None
    response_body_schema = None

    instance = None

    def handle_delete(self, *args, **kwargs):
        self.handle_kwargs(self.delete_kwargs_schema)
        self.handle_args(self.delete_args_schema)
        self.instance = self.get_instance()
        self.instance = self.delete()
        self.db.session.commit()
        return self.get_delete_response()

    def get_instance(self):
        return self.loaded_kwargs

    def delete(self):
        if is_sa_mapped(self.instance):
            self.db.session.delete(self.instance)
        return self.instance

    def get_delete_response(self):
        return self.get_schema_success_data(
            schema=self.delete_response_body_schema,
            data=self.instance,
        )



import setuptools


setuptools.setup(
    name="Flask-API-Framework",
    version="0.0.3",
    keywords="flask api framework",
    description="Flask API Framework",
    long_description="Please see the project links.",
    project_urls={
        "Documentation": "https://flask-api-framework.readthedocs.io/",
        "Source": "https://github.com/thnee/flask-api-framework",
    },
    license="BSD-3-Clause",
    author="Mattias Lindvall",
    author_email="mattias.lindvall@gmail.com",
    package_dir={"": "src"},
    packages=["flask_api_framework"],
    python_requires=">=3.7",
    install_requires=[
        "flask >= 2.0.0",
        "marshmallow >= 3.14.0",
    ],
    extras_require={
        "test": [
            "flask-sqlalchemy ~= 2.5.1",
            "marshmallow-sqlalchemy ~= 0.28.0",
            "flask-marshmallow ~= 0.14.0",
            "pytest ~= 7.1.2",
            "pytest-cov ~= 3.0.0",
            "flake8 ~= 4.0.1",
            "flake8-print ~= 5.0.0",
            "flake8-bugbear ~= 22.4.25",
            "black ~= 22.3.0",
            "isort ~= 5.10.1",
            "invoke ~= 1.7.1",
            "tox ~= 3.25.0",
        ],
        "build": [
            "wheel ~= 0.37.1",
            "twine ~= 4.0.0",
        ],
    },
    classifiers=[
        # "Development Status :: 1 - Planning",
        # "Development Status :: 2 - Pre-Alpha",
        "Development Status :: 3 - Alpha",
        # "Development Status :: 4 - Beta",
        # "Development Status :: 5 - Production/Stable",
        # "Development Status :: 6 - Mature",
        # "Development Status :: 7 - Inactive",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Software Development",
        "Intended Audience :: Developers",
        "Framework :: Flask",
    ],
)



import pytest


@pytest.mark.parametrize("examples_client", [["minimal_nodb", "app"]], indirect=True)
def test_index(examples_client):
    r = examples_client.get("/")
    assert r.json["value"] == "minimal"
